using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.WindowsMobile.DirectX;
using Microsoft.WindowsMobile.DirectX.Direct3D;
using GameUtils;

namespace Snake3D
{
    public partial class Snake3DForm : Form
    {
        #region Drawing variables

        /// <summary>
        /// Direct3D device we are going to draw on.
        /// </summary>
        Device device;

        #endregion

        #region Game Objects

        /// <summary>
        /// Keeps track of the position of a segment of the snake.
        /// </summary>
        class SnakeColoredBlock
        {
            /// <summary>
            /// Speed of the segment in the X direction.
            /// </summary>
            public float XSpeed;

            /// <summary>
            /// Speed of the segment in the X direction.
            /// </summary>
            public float YSpeed;

            /// <summary>
            /// Speed of the segment in the Z direction
            /// </summary>
            public float ZSpeed;

            /// <summary>
            /// Position of the segment
            /// </summary>
            public Matrix PositionMatrix;

            /// <summary>
            /// Size of the segment
            /// </summary>
            public Matrix ScaleMatrix;

            /// <summary>
            /// The yaw value for our segment to animate the tumble
            /// </summary>
            public float Yaw;

            /// <summary>
            /// Change value for the yaw. Set at random when the
            /// segment is constructed.
            /// </summary>
            public float YawSpeed;


            /// <summary>
            /// The pitch value for our segment to animate the tumble
            /// </summary>
            public float Pitch;

            /// <summary>
            /// Change value for the pitch. Set at random when the
            /// segment is constructed.
            /// </summary>
            public float PitchSpeed;

            /// <summary>
            /// The roll value for our segment to animate the tumble
            /// </summary>
            public float Roll;

            /// <summary>
            /// Change value for the roll. Set at random when the
            /// segment is constructed.
            /// </summary>
            public float RollSpeed;

            /// <summary>
            /// This defines the shape of our snake segments and the colour of the 
            /// vertices.
            /// </summary>
            public static CustomVertex.PositionColored[] Vertices;

            /// <summary>
            /// Vertex buffer which is used by the display hardware 
            /// recieve details of what we want to draw.
            /// </summary>
            public static VertexBuffer VertBuffer = null;

            /// <summary>
            /// Creates the single cube which is used by all the coloured blocks. Called 
            /// when the class is loaded.
            /// </summary>
            static SnakeColoredBlock()
            {
                Vertices = new CustomVertex.PositionColored[36];

                int colour = Color.Red.ToArgb();

                /// Front face
                Vertices[0] = new CustomVertex.PositionColored(-0.5f, -0.5f, -0.5f, colour);
                Vertices[1] = new CustomVertex.PositionColored(-0.5f, +0.5f, -0.5f, colour);
                Vertices[2] = new CustomVertex.PositionColored(+0.5f, -0.5f, -0.5f, colour);

                Vertices[3] = new CustomVertex.PositionColored(+0.5f, -0.5f, -0.5f, colour);
                Vertices[4] = new CustomVertex.PositionColored(-0.5f, +0.5f, -0.5f, colour);
                Vertices[5] = new CustomVertex.PositionColored(+0.5f, +0.5f, -0.5f, colour);

                colour = Color.Green.ToArgb();

                /// Right hand face
                Vertices[6] = new CustomVertex.PositionColored(+0.5f, -0.5f, -0.5f, colour);
                Vertices[7] = new CustomVertex.PositionColored(+0.5f, +0.5f, -0.5f, colour);
                Vertices[8] = new CustomVertex.PositionColored(+0.5f, -0.5f, +0.5f, colour);

                Vertices[9] = new CustomVertex.PositionColored(+0.5f, +0.5f, -0.5f, colour);
                Vertices[10] = new CustomVertex.PositionColored(+0.5f, +0.5f, +0.5f, colour);
                Vertices[11] = new CustomVertex.PositionColored(+0.5f, -0.5f, +0.5f, colour);


                colour = Color.Yellow.ToArgb();

                /// Top face
                Vertices[12] = new CustomVertex.PositionColored(-0.5f, -0.5f, -0.5f, colour);
                Vertices[13] = new CustomVertex.PositionColored(+0.5f, -0.5f, -0.5f, colour);
                Vertices[14] = new CustomVertex.PositionColored(+0.5f, -0.5f, +0.5f, colour);

                Vertices[15] = new CustomVertex.PositionColored(-0.5f, -0.5f, -0.5f, colour);
                Vertices[16] = new CustomVertex.PositionColored(+0.5f, -0.5f, +0.5f, colour);
                Vertices[17] = new CustomVertex.PositionColored(-0.5f, -0.5f, +0.5f, colour);

                colour = Color.Blue.ToArgb();

                /// left hand face
                Vertices[18] = new CustomVertex.PositionColored(-0.5f, -0.5f, -0.5f, colour);
                Vertices[19] = new CustomVertex.PositionColored(-0.5f, -0.5f, +0.5f, colour);
                Vertices[20] = new CustomVertex.PositionColored(-0.5f, +0.5f, -0.5f, colour);

                Vertices[21] = new CustomVertex.PositionColored(-0.5f, +0.5f, -0.5f, colour);
                Vertices[22] = new CustomVertex.PositionColored(-0.5f, -0.5f, +0.5f, colour);
                Vertices[23] = new CustomVertex.PositionColored(-0.5f, +0.5f, +0.5f, colour);

                colour = Color.Orange.ToArgb();

                /// Back face
                Vertices[24] = new CustomVertex.PositionColored(-0.5f, -0.5f, +0.5f, colour);
                Vertices[25] = new CustomVertex.PositionColored(+0.5f, -0.5f, +0.5f, colour);
                Vertices[26] = new CustomVertex.PositionColored(-0.5f, +0.5f, +0.5f, colour);

                Vertices[27] = new CustomVertex.PositionColored(+0.5f, -0.5f, +0.5f, colour);
                Vertices[28] = new CustomVertex.PositionColored(+0.5f, +0.5f, +0.5f, colour);
                Vertices[29] = new CustomVertex.PositionColored(-0.5f, +0.5f, +0.5f, colour);

                colour = Color.Indigo.ToArgb();

                /// Bottom face
                Vertices[30] = new CustomVertex.PositionColored(-0.5f, +0.5f, -0.5f, colour);
                Vertices[31] = new CustomVertex.PositionColored(-0.5f, +0.5f, +0.5f, colour);
                Vertices[32] = new CustomVertex.PositionColored(+0.5f, +0.5f, +0.5f, colour);

                Vertices[33] = new CustomVertex.PositionColored(-0.5f, +0.5f, -0.5f, colour);
                Vertices[34] = new CustomVertex.PositionColored(+0.5f, +0.5f, +0.5f, colour);
                Vertices[35] = new CustomVertex.PositionColored(+0.5f, +0.5f, -0.5f, colour);
            }

            /// <summary>
            /// Create a segment
            /// </summary>
            /// <param name="device">drawing device</param>
            public SnakeColoredBlock(float width, float height, float depth, Device device, Random rand)
            {
                PositionMatrix = Matrix.Translation(0.0f, 0.0f, 0.0f);
                ScaleMatrix = Matrix.Scaling(width, height, depth);

                RollSpeed = rand.Next(300) / 1000.0f;
                PitchSpeed = rand.Next(300) / 1000.0f;
                YawSpeed = rand.Next(300) / 1000.0f;

                /// Create a vertex buffer for our segment
                /// We can't do this on construction because we need a device
                /// reference to create the vertex buffer.

                if (VertBuffer == null)
                {
                    VertBuffer = new VertexBuffer(
                        typeof(CustomVertex.PositionColored),   // type of the buffer
                        Vertices.Length,                        // holding vertices array
                        device,                                 // for our device
                        Usage.WriteOnly,                        // never going to read it
                        CustomVertex.PositionColored.Format,    // source format
                        Pool.SystemMemory);                     // mobile 3D requires this
                }
            }

            /// <summary>
            /// Draw the segment
            /// </summary>
            /// <param name="device">destination device for the draw</param>
            public void Draw(Device device)
            {
                // Set the data in the vertex buffer to our triangles
                VertBuffer.SetData(Vertices, 0, LockFlags.None);

                Yaw += YawSpeed;
                Pitch += PitchSpeed;
                Roll += RollSpeed;

                device.Transform.World = ScaleMatrix * Matrix.RotationYawPitchRoll(Yaw, Pitch, Roll) * PositionMatrix ;

                // Point the device at our vertex buffer
                device.SetStreamSource(0, VertBuffer, 0);

                // Ask the device to draw the contents of the buffer
                device.DrawPrimitives(PrimitiveType.TriangleList, 0, 12);
            }

            /// <summary>
            /// Check to see if this segment intersects with 
            /// the bonus square.
            /// </summary>
            /// <param name="bonus">bonus we are looking for</param>
            /// <returns>true if th</returns>
            public bool IntersectsWith(Bonus bonus)
            {
                float x = PositionMatrix.M41;
                float y = PositionMatrix.M42;
                float xe = x + ScaleMatrix.M11;
                float ye = y + ScaleMatrix.M22;

                float bX = bonus.PositionMatrix.M41;
                float bY = bonus.PositionMatrix.M42;
                float bXe = bX + bonus.ScaleMatrix.M11;
                float bYe = bY + bonus.ScaleMatrix.M22;

                if (x > bX && x < bXe && y > bY && y < bYe)
                {
                    return true;
                }

                if (xe > bX && xe < bXe && ye > bY && ye < bYe)
                {
                    return true;
                }

                if (xe > bX && xe < bXe && y > bY && y < bYe)
                {
                    return true;
                }
                if (x > bX && x < bXe && ye > bY && ye < bYe)
                {
                    return true;
                }
                return false;
            }
        }

        class Snake
        {
            /// <summary>
            /// X coordinate of top left hand corner of the
            /// snake head. Origin at centre.
            /// </summary>
            public float X;

            /// <summary>
            /// Y coordinate of top left hand corner of the 
            /// snake head. Origin at centre.
            /// </summary>
            public float Y;

            /// <summary>
            /// Z coordinate of top left hand corner of the snake
            /// head. Increasing Z moves into the picture. 
            /// </summary>
            public float Z;

            /// <summary>
            /// The longest the snake is allowed to get in segments.
            /// </summary>
            private int maxSnakeLength;

            /// <summary>
            /// The number of updates between each grow action.
            /// </summary>
            private int growSpeed;

            /// <summary>
            /// Counter for snake growing. As the time goes by the snake
            /// will grow. growCount is updated each time the snake is
            /// updated. When it reaches growSpeed the snake length
            /// is increased.
            /// </summary>
            private int growCount;

            /// <summary>
            /// The length of the snake at the start of the game.
            /// </summary>
            public int startSnakeLength;

            /// <summary>
            /// The current length of the snake in segments.
            /// </summary>
            public int currentSnakeLength = 0;

            /// <summary>
            /// The entire snake. Set up when the program is
            /// initialised. Use increasing amounts of the snake
            /// as the game progresses.
            /// </summary>
            private SnakeColoredBlock[] wholeSnake;

            /// <summary>
            /// Draw the snake on the graphics surface given.
            /// </summary>
            /// <param name="device">Device used to draw on</param>
            public void Draw(Device device)
            {
                for (int segNo = 0; segNo < currentSnakeLength; segNo++)
                {
                    wholeSnake[segNo].Draw(device);
                }
            }

            /// <summary>
            /// Updates the snake.
            /// </summary>
            /// <param name="xSpeed">The X distance to move the snake head.</param>
            /// <param name="ySpeed">The Y distance to move the snake head/</param>
            /// <param name="zSpeed">The Z distance to move the snake head</param>
            /// <returns>True if the update succeeded. 
            /// False if the snake crashes into itself.</returns>
            public bool Update(float xSpeed, float ySpeed, float zSpeed)
            {

                if (currentSnakeLength < startSnakeLength)
                {
                    wholeSnake[currentSnakeLength].XSpeed = 0.0f;
                    wholeSnake[currentSnakeLength].YSpeed = 0.0f;
                    wholeSnake[currentSnakeLength].ZSpeed = 0.0f;
                    currentSnakeLength++;
                }
                else
                {
                    if (++growCount >= growSpeed)
                    {
                        growCount = 0;
                        if (currentSnakeLength < maxSnakeLength)
                        {
                            wholeSnake[currentSnakeLength].XSpeed = 0.0f;
                            wholeSnake[currentSnakeLength].YSpeed = 0.0f;
                            wholeSnake[currentSnakeLength].ZSpeed = 0.0f;
                            currentSnakeLength++;
                        }
                    }
                }

                X += xSpeed;
                Y += ySpeed;
                Z += zSpeed;

                SnakeColoredBlock head = wholeSnake[0];

                for (int segNo = currentSnakeLength - 1; segNo > 0; segNo--)
                {
                    SnakeColoredBlock segment = wholeSnake[segNo];
                    SnakeColoredBlock segmentInFront = wholeSnake[segNo - 1];
                    segmentInFront = wholeSnake[segNo - 1];

                    segment.PositionMatrix = segmentInFront.PositionMatrix;

                    //segment.Position.M41 = segmentInFront.Position.M41;
                    //segment.Position.M42 = segmentInFront.Position.M42;
                    //segment.Position.M43 = segmentInFront.Position.M43;

                    segment.XSpeed = segmentInFront.XSpeed;
                    segment.YSpeed = segmentInFront.YSpeed;
                    segment.ZSpeed = segmentInFront.ZSpeed;
                }

                head.PositionMatrix = Matrix.Translation(X, Y, Z);

                /// This is a faster way of updating a translation matrix. I happen to know that
                /// the values in the matrix which specify the translation to be applied are specified
                /// in locations M41, M42 and M43. I can therefore use this code instead of the above
                /// to slightly increase the speed of the drawing. I could also use it above
                /// to reduce the amount of work that is performed when assigning one translation
                /// matrix to another.
                /// 
                //head.Position.M41 = X;
                //head.Position.M42 = Y;
                //head.Position.M43 = Z;

                head.XSpeed = xSpeed;
                head.YSpeed = ySpeed;
                head.ZSpeed = zSpeed;

                return true;
            }

            /// <summary>
            /// Used to see if any part of the snake intersects with any a given part of 
            /// the playfield. Used for placing bonus items or detecting hits.
            /// </summary>
            /// <param name="target">the rectangle to be tested</param>
            /// <returns>true if any part of the snake intersects with the rectangle</returns>
            public bool IntersectsWith(Bonus bonus)
            {
                foreach (SnakeColoredBlock segment in wholeSnake)
                {
                    if (segment.IntersectsWith(bonus))
                    {
                        return true;
                    }
                }
                return false;
            }

            /// <summary>
            /// Resets the snake for the next game. Places the snake on the
            /// screen at the given position and makes it zero length. It
            /// will grow to the start length as it is redrawn.
            /// </summary>
            /// <param name="inSnakeX">start X position</param>
            /// <param name="inSnakeY">start Y position.</param>
            /// <param name="inSnakeZ">start Z position.</param>
            /// <param name="inStartSnakeLength">length to grow to at the start.</param>
            /// <param name="inMaxSnakeLength">the maximum length the snake is to have</param>
            /// <param name="inGrowSpeed">the number of updates between grows</param>
            public void ResetSnake(
                float inSnakeX, float inSnakeY, float inSnakeZ,
                int inStartSnakeLength,
                int inMaxSnakeLength,
                int inGrowSpeed)
            {
                X = inSnakeX;
                Y = inSnakeY;
                Z = inSnakeZ;
                startSnakeLength = inStartSnakeLength;
                maxSnakeLength = inMaxSnakeLength;
                growSpeed = inGrowSpeed;
                growCount = 0;
                currentSnakeLength = 0;
            }

            public Snake(int inMaxSnakeLength, 
                float segmentWidth, float segmentHeight, float segmentDepth,
                Device device,
                Random rand)
            {
                maxSnakeLength = inMaxSnakeLength;

                wholeSnake = new SnakeColoredBlock[maxSnakeLength];

                wholeSnake[0] = new SnakeColoredBlock( segmentWidth, segmentHeight, segmentDepth,device, rand);

                for (int segNo = 1; segNo < maxSnakeLength; segNo++)
                {
                    wholeSnake[segNo] = new SnakeColoredBlock( segmentWidth, segmentHeight, segmentDepth, device, rand);
                }
            }
        }

        public class ImageBackground
        {

            /// <summary>
            /// Texture to be drawn with. Passed into the constructor.
            /// </summary>
            public Texture BackgroundTexture;

            /// <summary>
            /// This defines the shape of our background image and the texture of the 
            /// vertices.
            /// </summary>
            public CustomVertex.PositionTextured[] Vertices;

            /// <summary>
            /// Vertex buffer which is used by the display hardware 
            /// recieve details of what we want to draw.
            /// </summary>
            public VertexBuffer VertBuffer = null;

            public ImageBackground(Device device, System.IO.Stream textureStream, float size, float zValue)
            {

                BackgroundTexture = TextureLoader.FromStream(device, textureStream);

                Vertices = new CustomVertex.PositionTextured[6];

                Vertices[0] = new CustomVertex.PositionTextured(-size, -size, zValue, 0.0f, 0.0f);
                Vertices[1] = new CustomVertex.PositionTextured(-size, +size, zValue, 0.0f, 1.0f);
                Vertices[2] = new CustomVertex.PositionTextured(+size, -size, zValue, 1.0f, 0.0f);

                Vertices[3] = new CustomVertex.PositionTextured(+size, -size, zValue, 1.0f, 0.0f);
                Vertices[4] = new CustomVertex.PositionTextured(-size, +size, zValue, 0.0f, 1.0f);
                Vertices[5] = new CustomVertex.PositionTextured(+size, +size, zValue, 1.0f, 1.0f);

                VertBuffer = new VertexBuffer(
                    typeof(CustomVertex.PositionTextured),   // type of the buffer
                    Vertices.Length,                        // holding vertices array
                    device,                                 // for our device
                    Usage.WriteOnly,                        // never going to read it
                    CustomVertex.PositionTextured.Format,    // source format
                    Pool.SystemMemory);                     // mobile 3D requires this

            }

            /// <summary>
            /// Draw the textured image.
            /// </summary>
            /// <param name="device">target device to draw on</param>
            public void Draw ( Device device )
            {
                // Set the data in the vertex buffer to our triangles
                VertBuffer.SetData(Vertices, 0, LockFlags.None);

                // clear off any existing transformations
                device.Transform.World = Matrix.Identity;

                // Point the device at our vertex buffer
                device.SetStreamSource(0, VertBuffer, 0);

                // Assign the texture to the device.
                device.SetTexture(0, BackgroundTexture); 

                // Ask the device to draw the contents of the buffer
                device.DrawPrimitives(PrimitiveType.TriangleList, 0, 2);

                device.SetTexture(0, null);

            }
        }

        public class PlainBackground
        {

            /// <summary>
            /// Texture to be drawn with. Passed into the constructor.
            /// </summary>
            public Texture BackgroundTexture;

            /// <summary>
            /// This defines the shape of our background image and the texture of the 
            /// vertices.
            /// </summary>
            public CustomVertex.PositionColored[] Vertices;

            /// <summary>
            /// Vertex buffer which is used by the display hardware 
            /// recieve details of what we want to draw.
            /// </summary>
            public VertexBuffer VertBuffer = null;


            public PlainBackground(Device device, Color backColor, float size, float zValue)
            {
                int color = backColor.ToArgb();

                Vertices = new CustomVertex.PositionColored[6];

                Vertices[0] = new CustomVertex.PositionColored(-size, -size, zValue, color);
                Vertices[1] = new CustomVertex.PositionColored(-size, +size, zValue, color);
                Vertices[2] = new CustomVertex.PositionColored(+size, -size, zValue, color);

                Vertices[3] = new CustomVertex.PositionColored(+size, -size, zValue, color);
                Vertices[4] = new CustomVertex.PositionColored(-size, +size, zValue, color);
                Vertices[5] = new CustomVertex.PositionColored(+size, +size, zValue, color);

                VertBuffer = new VertexBuffer(
                    typeof(CustomVertex.PositionColored),   // type of the buffer
                    Vertices.Length,                        // holding vertices array
                    device,                                 // for our device
                    Usage.WriteOnly,                        // never going to read it
                    CustomVertex.PositionColored.Format,    // source format
                    Pool.SystemMemory);                     // mobile 3D requires this

            }

            public void Draw(Device device)
            {
                // Set the data in the vertex buffer to our triangles
                VertBuffer.SetData(Vertices, 0, LockFlags.None);

                // clear off any existing transformation
                device.Transform.World = Matrix.Identity;

                // Point the device at our vertex buffer
                device.SetStreamSource(0, VertBuffer, 0);

                // Ask the device to draw the contents of the buffer
                device.DrawPrimitives(PrimitiveType.TriangleList, 0, 2);
            }
        }


        #endregion


        #region Bonus

        #region Bonus class

        /// <summary>
        /// Holds information about a bonus item on the screen
        /// </summary>
        class Bonus
        {
            /// <summary>
            /// True if the bonus is presently enabled
            /// </summary>
            public bool Enabled;

            /// <summary>
            /// This defines two triangles, centred about the origin, with the required size.
            /// </summary>
            public CustomVertex.PositionColored[] Vertices;

            /// <summary>
            /// Vertex buffer which is used by the display hardware 
            /// recieve details of what we want to draw.
            /// </summary>
            public VertexBuffer VertBuffer;

            /// <summary>
            /// Position of the segment
            /// </summary>
            public Matrix PositionMatrix;

            /// <summary>
            /// Scale to draw the segment
            /// </summary>
            public Matrix ScaleMatrix;

            /// <summary>
            /// Value gained when the bonus is eaten.
            /// </summary>
            public int Value;

            /// <summary>
            /// Random number generator to use to place food.
            /// </summary>
            public Random Rand;

            /// <summary>
            /// Number of game ticks that this bonus has to live
            /// </summary>
            public int LifeLength;

            /// <summary>
            /// Counter for bonus life
            /// </summary>
            public int LifeCount;

            /// <summary>
            /// Create a new bonus item.
            /// </summary>
            /// <param name="inDiameter">Size of the bonus.</param>
            /// <param name="inBonusBrush">Brush to draw it with.</param>
            /// <param name="inValue">The amount the bonus is worth</param>
            /// <param name="inRand">The random number generator to use.</param>
            public Bonus(float bonusWidth, float bonusHeight, int inValue, Color bonusColor, Random inRand, Device device)
            {
                int colour = bonusColor.ToArgb();

                Vertices = new CustomVertex.PositionColored[6];

                Vertices[0] = new CustomVertex.PositionColored(-0.5f, -0.5f, 0, colour);
                Vertices[1] = new CustomVertex.PositionColored(-0.5f, +0.5f, 0, colour);
                Vertices[2] = new CustomVertex.PositionColored(+0.5f, -0.5f, 0, colour);

                Vertices[3] = new CustomVertex.PositionColored(+0.5f, -0.5f, 0, colour);
                Vertices[4] = new CustomVertex.PositionColored(-0.5f, +0.5f, 0, colour);
                Vertices[5] = new CustomVertex.PositionColored(+0.5f, +0.5f, 0, colour);

                /// Create a vertex buffer for our segment
                VertBuffer = new VertexBuffer(
                    typeof(CustomVertex.PositionColored),   // type of the buffer
                    6,                                      // holding 3 vertices
                    device,                                 // for our device
                    Usage.WriteOnly,                        // never going to read it
                    CustomVertex.PositionColored.Format,    // source format
                    Pool.SystemMemory);                     // mobile 3D requires this

                ScaleMatrix = Matrix.Scaling(bonusWidth, bonusHeight, 1);
                PositionMatrix = Matrix.Translation(0, 0, 0);
                Value = inValue;
                Rand = inRand;
            }

            /// <summary>
            /// Places the bonus on the screen in a place which does not
            /// intersect with the given snake.
            /// </summary>
            /// <param name="percentLikely">The percentage likeliness of gettting a 
            /// bonus. The bigger the number, up to 100, the likelier the 
            /// bonus.</param>
            /// <param name="snake">The snake we must not intersect with</param>
            /// <returns>true if the placing worked. False if not.</returns>
            /// <param name="maxX">Maximum X value for the position</param>
            /// <param name="maxY">Maximum Y value for the position</param>
            public bool PlaceBonus(int percentLikely, Snake snake)
            {
                /// Don't place the bonus if it is already there
                if (Enabled)
                {
                    return true;
                }

                /// Only appear every now and then
                if (Rand.Next(100) > percentLikely)
                {
                    return false;
                }

                /// Try 5 times.
                for (int i = 0; i < 5; i++)
                {
                    PositionMatrix.M41 = (Rand.Next(20) / 10.0f) - 1.0f;
                    PositionMatrix.M42 = (Rand.Next(20) / 10.0f) - 1.0f;
                    if (!snake.IntersectsWith(this))
                    {
                        LifeLength = Rand.Next(50) + 20;
                        LifeCount = 0;
                        Enabled = true;
                        return true;
                    }
                }
                return false;
            }

            /// <summary>
            /// Check to see if the bonus and the snake intersect.
            /// </summary>
            /// <param name="snake">snake to check against</param>
            /// <returns>true if the bonus has been won</returns>
            public bool CheckBonus(Snake snake)
            {
                if (snake.IntersectsWith(this))
                {
                    
                    Enabled = false;
                    return true;
                }
                else
                {
                    if (LifeCount++ >= LifeLength)
                    {
                        Enabled = false;
                        return false;
                    }
                }
                return false;
            }

            /// <summary>
            /// Draw the segment
            /// </summary>
            /// <param name="device">destination device for the draw</param>
            public void Draw(Device device)
            {
                // Set the data in the vertex buffer to our triangles
                VertBuffer.SetData(Vertices, 0, LockFlags.None);

                /// Position the segment in
                device.Transform.World = ScaleMatrix * PositionMatrix;

                // Point the device at our vertex buffer
                device.SetStreamSource(0, VertBuffer, 0);

                // Ask the device to draw the contents of the buffer
                device.DrawPrimitives(PrimitiveType.TriangleList, 0, 2);
            }

        }

        #endregion

        #region Bonus Management

        /// <summary>
        /// The one and only bonus value.
        /// </summary>
        Bonus bonus;

        #endregion

        #endregion


        #region Game Sound Effects

        /// <summary>
        /// sound made on each snake move
        /// </summary>
        private Sound clickSound;

        /// <summary>
        /// Sound made when the snake eats something.
        /// </summary>
        private Sound eatSound;

        #endregion 

        public Snake3DForm()
        {
            InitializeComponent();
        }

        #region Timer routines

        /// <summary>
        /// The tick count recorded at the start of the timing period.
        /// </summary>
        private int clockStartTick;

        /// <summary>
        /// Counter for the number of readings to be taken.
        /// </summary>
        private int readingCount = 0;

        /// <summary>
        /// Number of readings to be averaged over.
        /// </summary>
        private int sampleSize = 20;

        /// <summary>
        /// Total of readings taken.
        /// </summary>
        private float readingTotal = 0.0f;

        /// <summary>
        /// Calculated average result.
        /// </summary>
        private float timerResult;

        /// <summary>
        /// Starts the timer ticking.
        /// </summary>
        private void startTimer()
        {
            clockStartTick = System.Environment.TickCount;
        }

        /// <summary>
        /// Obtains a new time value, adds it to the total and
        /// displays the average if we are at the end of a sample
        /// period.
        /// </summary>
        private void showTimer()
        {
            readingTotal += (System.Environment.TickCount - clockStartTick);

            if (++readingCount == sampleSize)
            {
                timerResult = readingTotal / sampleSize;
                this.Text = "Time : " + timerResult.ToString();
                readingTotal = 0.0f;
                readingCount = 0;
            }
        }

        #endregion

        #region Game Management

        #region Game State variables

        /// <summary>
        /// When true the game is being played.
        /// </summary>
        private bool gameActive = false;

        /// <summary>
        /// When true the game is in pause state.
        /// </summary>
        private bool gamePaused = false;

        #endregion

        #region Snake Drawing variables

        /// <summary>
        /// The snake in our game.
        /// </summary>
        private Snake snake;

        /// <summary>
        /// The background image in our game.
        /// </summary>
        private ImageBackground imageBackground;

        /// <summary>
        /// Plain version of the image
        /// </summary>
        private PlainBackground plainBackground;

        /// <summary>
        /// Currently executing assembly. Cached because it is used
        /// to load a lot of resources. 
        /// </summary>
        readonly System.Reflection.Assembly assembly =
            System.Reflection.Assembly.GetExecutingAssembly();

        #endregion 


        #region Snake movement variables

        /// <summary>
        /// Speed of snake in x direction.
        /// </summary>
        private float snakeXSpeed;

        /// <summary>
        /// Speed of snake in y direction.
        /// </summary>
        private float snakeYSpeed;

        /// <summary>
        /// The speed of movement of the snake itself.
        /// </summary>
        private float snakeSpeed;

        #endregion


        PresentParameters presentParams;

        public void Init()
        {
            presentParams = new PresentParameters();
            presentParams.Windowed = true;
            presentParams.SwapEffect = SwapEffect.Discard;
            presentParams.EnableAutoDepthStencil = true;
            presentParams.AutoDepthStencilFormat = DepthFormat.D16;

            device = new Device(
                0,                      // device number 0
                DeviceType.Default,     // default configuration
                this,                   // reference to the parent window
                CreateFlags.None,       // no special creation flags
                presentParams);         // the presentation parameters

            snake = new Snake(30, .2f, .2f, .2f, device, new Random());

            bonus = new Bonus(0.5f, 0.5f, 50, Color.Yellow, new System.Random(1), device);

            imageBackground = new ImageBackground(device,
                assembly.GetManifestResourceStream("Snake3D.images.grass.jpg"),
                2.0f,
                0.2f);

            clickSound = new Sound
                (assembly.GetManifestResourceStream("Snake3D.sounds.click.wav"));
            eatSound = new Sound
                (assembly.GetManifestResourceStream("Snake3D.sounds.burp.wav"));

            plainBackground = new PlainBackground(device, Color.PaleGreen, 3.0f, 0.2f);

            device.DeviceReset += new EventHandler(DeviceReset);

            DeviceReset(device, EventArgs.Empty);
        }

        /// <summary>
        /// Ratio for the currently active display. Used when changing the viewing
        /// position.
        /// </summary>
        private float aspectRatio;

        void DeviceReset(System.Object sender, System.EventArgs e ) {

            device.RenderState.Lighting = false;
            
            /// Controls whether or not we will draw triangles which are "not visible", i.e.
            /// whose draw order relative to the camera is clockwise. Turn this on (i.e. use
            /// the default) if you are drawing in 3D and you have designed your objects so that
            /// the draw order is correct (take a look at the SnakeColoredBlock class for 
            /// an example of this.
            /// 
            /// Turn this off if your objects are 2D and you want to view them 
            /// from the back as well as the front If you only ever view from one side you 
            /// should still turn it off when drawing in 2D because it speeds things up slightly.
            /// 

            // device.RenderState.CullMode = Cull.None;

            device.Transform.View = Matrix.LookAtLH(
                new Vector3(0.0f, 0.0f, -5.0f), // camera position
                new Vector3(0.0f, 0.0f, 0.0f),  // camera target
                new Vector3(0.0f, 1.0f, 0.0f)   // camera up vector
                );

            aspectRatio =
                ((float)device.PresentationParameters.BackBufferWidth) /
                device.PresentationParameters.BackBufferHeight;

            device.Transform.Projection = Matrix.PerspectiveFovLH(
                (float)Math.PI / 4,     // field of view
                aspectRatio,                   // aspect ratio
                1.0f,                   // near Z plane
                100.0f                  // far Z plane
                );

        }

        private void Render()
        {
            device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, Color.CornflowerBlue, 1.0f, 0);

            device.BeginScene();

            //imageBackground.Draw(device);
            plainBackground.Draw(device);
            snake.Draw(device);

            bonus.Draw(device);

            device.EndScene();
            device.Present();
        }


        /// <summary>
        /// Plays the game. Sets gameActive to true and
        /// updates the game until it becomes false.
        /// </summary>
        private void playGame()
        {
            gameActive = true;
            while (gameActive)
            {
                startTimer();
                if (!gamePaused)
                {
                    snake.Update(snakeXSpeed, snakeYSpeed, 0.0f);
                    if (bonus.Enabled)
                    {
                        if (bonus.CheckBonus(snake))
                        {
                            eatSound.Play();
                        }
                    }
                    else
                    {
                        bonus.PlaceBonus(80, snake);
                    }
                }
                Application.DoEvents();
                Render();
                showTimer();
                System.Threading.Thread.Sleep(50);
            }
        }

        /// <summary>
        /// Refers to a method with the signature required for 
        /// Form.Invoke calls.
        /// </summary>
        private delegate void gameMethod();

        /// <summary>
        /// Refers to the method which will play the game. Set
        /// to playGame when the game starts.
        /// </summary>
        gameMethod playDelegate;

        /// <summary>
        /// Starts a new game.
        /// </summary>
        private void startGame()
        {
            bonus.Rand = new Random(1);
            snake.ResetSnake(0, 0, 0, 1, 30, 6);
            snakeSpeed = 0.4f;

            snakeXSpeed = snakeSpeed;
            snakeYSpeed = 0;

            playDelegate = new gameMethod(playGame);
            Invoke(playDelegate);
        }

        private void doPause()
        {
            gamePaused = !gamePaused;
        }

        #endregion


        #region Event Handlers

        protected override void OnPaintBackground(PaintEventArgs e)
        {
        }

        protected override void OnPaint(PaintEventArgs e)
        {
            Render();
        }

        #endregion 

        private void playMenuItem_Click(object sender, EventArgs e)
        {
            startGame();
        }

        private void Snake3DForm_KeyDown(object sender, KeyEventArgs e)
        {
            if ((e.KeyCode == System.Windows.Forms.Keys.F1))
            {
                // Soft Key 1
                // Not handled when menu is present.
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F2))
            {
                // Soft Key 2
                // Not handled when menu is present.
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Up))
            {
                // Up
                snakeXSpeed = 0.0f;
                snakeYSpeed = snakeSpeed;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Down))
            {
                // Down
                snakeXSpeed = 0.0f;
                snakeYSpeed = -snakeSpeed;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Left))
            {
                // Left
                snakeXSpeed = -snakeSpeed;
                snakeYSpeed = 0.0f;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Right))
            {
                // Right
                snakeXSpeed = snakeSpeed;
                snakeYSpeed = 0.0f;
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.Enter))
            {
                // Enter
                doPause();
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D1))
            {
                // 1
                // 1
                device.Transform.View = Matrix.LookAtLH(
                    new Vector3(0.0f, 0.0f, -5.0f), // camera position
                    new Vector3(0.0f, 0.0f, 0.0f),  // camera target
                    new Vector3(0.0f, 1.0f, 0.0f)   // camera up vector
                    );

                device.Transform.Projection = Matrix.PerspectiveFovLH(
                    (float)Math.PI / 4,     // field of view
                    1.0f,                   // aspect ratio
                    1.0f,                   // near Z plane
                    100.0f                  // far Z plane
                    );
                Render();
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D2))
            {
                // 2
                device.Transform.View = Matrix.LookAtLH(
                    new Vector3(0.0f, 0.0f, -2.0f), // camera position
                    new Vector3(0.0f, 0.0f, 0.0f),  // camera target
                    new Vector3(0.0f, 1.0f, 0.0f)   // camera up vector
                    );
                device.Transform.Projection = Matrix.PerspectiveFovLH(
                    (float)Math.PI / 4,     // field of view
                    aspectRatio,            // aspect ratio
                    1.0f,                   // near Z plane
                    100.0f                  // far Z plane
                    );
                Render();
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D3))
            {
                // 3
                device.Transform.View = Matrix.LookAtLH(
                    new Vector3(0.0f, 0.0f, -5.0f), // camera position
                    new Vector3(0.0f, 0.0f, 0.0f),  // camera target
                    new Vector3(1.0f, 1.0f, 0.0f)   // camera up vector
                    );
                device.Transform.Projection = Matrix.PerspectiveFovLH(
                    (float)Math.PI / 4,     // field of view
                    aspectRatio,            // aspect ratio
                    1.0f,                   // near Z plane
                    100.0f                  // far Z plane
                    );
                Render();
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D4))
            {
                // 4
                device.Transform.View = Matrix.LookAtLH(
                    new Vector3(0.0f, -1.0f, -1.0f), // camera position
                    new Vector3(0.0f, 0.0f, 0.0f),  // camera target
                    new Vector3(0.0f, 1.0f, 0.0f)   // camera up vector
                    );

                device.Transform.Projection = Matrix.PerspectiveFovLH(
                    (float)Math.PI / 2,     // field of view
                    aspectRatio,            // aspect ratio
                    1.0f,                   // near Z plane
                    100.0f                  // far Z plane
                    );
                Render();
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D5))
            {
                // 5
                device.Transform.View = Matrix.LookAtLH(
                    new Vector3(0.0f, 0.0f, 2.0f), // camera position
                    new Vector3(0.0f, 0.0f, 0.0f),  // camera target
                    new Vector3(0.0f, 1.0f, 0.0f)   // camera up vector
                    );
                device.Transform.Projection = Matrix.PerspectiveFovLH(
                    (float)Math.PI / 2,     // field of view
                    aspectRatio,            // aspect ratio
                    1.0f,                   // near Z plane
                    100.0f                  // far Z plane
                    );
                Render();

            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D6))
            {
                // 6
                snake.ResetSnake(0, 0, 0, 2, 5, 1);
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D7))
            {
                // 7
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D8))
            {
                // 8
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D9))
            {
                // 9
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F8))
            {
                // *
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.D0))
            {
                // 0
            }
            if ((e.KeyCode == System.Windows.Forms.Keys.F9))
            {
                // #
            }

        }

        private void pauseMenuItem_Click(object sender, EventArgs e)
        {
            doPause();
        }
    }
}